#include <benchmark/benchmark.h>
#include "plf_colony.h"
#include <vector>
#include <deque>


// Fast xorshift+128 random number generator function (original: https://codingforspeed.com/using-faster-psudo-random-generator-xorshift/)
unsigned int xor_rand()
{
	static unsigned int x = 123456789;
	static unsigned int y = 362436069;
	static unsigned int z = 521288629;
	static unsigned int w = 88675123;

	const unsigned int t = x ^ (x << 11);

	// Rotate the static values (w rotation in return statement):
	x = y;
	y = z;
	z = w;

	return w = w ^ (w >> 19) ^ (t ^ (t >> 8));
}



double total = 0; // Used to stop compiler optimizing out functions in creation and iteration benchmarks




static void benchmark_colony_creation(benchmark::State& state) 
{
	for (auto _ : state)
	{
		plf::colony<int> test;
		total += test.size(); // To avoid compiler optimizing out creation of object
	}
}


static void benchmark_deque_creation(benchmark::State& state) 
{
	for (auto _ : state)
	{
		std::deque<int> test;
		total += test.size();
	}
}


static void benchmark_vector_creation(benchmark::State& state) 
{
	for (auto _ : state)
	{
		std::vector<int> test;
		total += test.size();
	}
}


// Register the function as a benchmark
BENCHMARK(benchmark_colony_creation);
BENCHMARK(benchmark_deque_creation);
BENCHMARK(benchmark_vector_creation);



static void benchmark_colony_insertion(benchmark::State& state) 
{
	for (auto _ : state)
	{
		plf::colony<int> test;
		
		for (unsigned int counter = 0; counter != 100000; ++counter)
		{
			test.insert(15);
		}
	}
}


static void benchmark_deque_insertion(benchmark::State& state) 
{
	for (auto _ : state)
	{
		std::deque<int> test;
		
		for (unsigned int counter = 0; counter != 100000; ++counter)
		{
			test.push_back(15);
		}
	}
}


static void benchmark_vector_insertion(benchmark::State& state) 
{
	for (auto _ : state)
	{
		std::vector<int> test;
		
		for (unsigned int counter = 0; counter != 100000; ++counter)
		{
			test.push_back(15);
		}
	}
}


// Register the function as a benchmark
BENCHMARK(benchmark_colony_insertion);
BENCHMARK(benchmark_deque_insertion);
BENCHMARK(benchmark_vector_insertion);



static void benchmark_colony_erasure(benchmark::State& state) 
{
	plf::colony<int> test;
	
	for (unsigned int counter = 0; counter != 100000; ++counter)
	{
		test.insert(15);
	}

	for (auto _ : state)
	{
		plf::colony<int> copy(test);
		
		for(plf::colony<int>::iterator it = copy.begin(); it != copy.end();)
		{
			if ((xor_rand() & 7) != 0)
			{
				++it;
			}
			else
			{
				it = copy.erase(it);
			}
		}
	}
}


static void benchmark_deque_erasure(benchmark::State& state) 
{
	std::deque<int> test;
	
	for (unsigned int counter = 0; counter != 100000; ++counter)
	{
		test.push_back(15);
	}

	for (auto _ : state)
	{
		std::deque<int> copy(test);
		
		for(std::deque<int>::iterator it = copy.begin(); it != copy.end();)
		{
			if ((xor_rand() & 7) != 0)
			{
				++it;
			}
			else
			{
				it = copy.erase(it);
			}
		}
	}
}



static void benchmark_deque_erasure_remove_if(benchmark::State& state) 
{
	std::deque<int> test;
	
	for (unsigned int counter = 0; counter != 100000; ++counter)
	{
		test.push_back(xor_rand() & 255);
	}

	for (auto _ : state)
	{
		std::deque<int> copy(test);
		
		copy.erase(std::remove_if(copy.begin(), copy.end(), [](int x){return x < 32;}), copy.end());
	}
}



static void benchmark_vector_erasure(benchmark::State& state) 
{
	std::vector<int> test;
	
	for (unsigned int counter = 0; counter != 100000; ++counter)
	{
		test.push_back(15);
	}

	for (auto _ : state)
	{
		std::vector<int> copy(test);
		
		for(std::vector<int>::iterator it = copy.begin(); it != copy.end();)
		{
			if ((xor_rand() & 7) != 0)
			{
				++it;
			}
			else
			{
				it = copy.erase(it);
			}
		}
	}
}


static void benchmark_vector_erasure_remove_if(benchmark::State& state) 
{
	std::vector<int> test;
	
	for (unsigned int counter = 0; counter != 100000; ++counter)
	{
		test.push_back(xor_rand() & 255);
	}

	for (auto _ : state)
	{
		std::vector<int> copy(test);
		
		copy.erase(std::remove_if(copy.begin(), copy.end(), [](int x){return x < 32;}), copy.end());
	}
}



// Register the function as a benchmark
BENCHMARK(benchmark_colony_erasure);
BENCHMARK(benchmark_deque_erasure);
BENCHMARK(benchmark_deque_erasure_remove_if);
BENCHMARK(benchmark_vector_erasure);
BENCHMARK(benchmark_vector_erasure_remove_if);




static void benchmark_colony_iterate_and_sum(benchmark::State& state) 
{
	plf::colony<int> test;
	
	for (unsigned int counter = 0; counter != 100000; ++counter)
	{
		test.insert(15);
	}

	for(plf::colony<int>::iterator it = test.begin(); it != test.end();)
	{
		if ((xor_rand() & 7) != 0)
		{
			++it;
		}
		else
		{
			it = test.erase(it);
		}
	}

	for (auto _ : state)
	{
		for(plf::colony<int>::iterator it = test.begin(); it != test.end(); ++it)
		{
			total += *it;
		}
	}
}



static void benchmark_deque_iterate_and_sum(benchmark::State& state) 
{
	std::deque<int> test;
	
	for (unsigned int counter = 0; counter != 100000; ++counter)
	{
		test.push_back(15);
	}

	for(std::deque<int>::iterator it = test.begin(); it != test.end();)
	{
		if ((xor_rand() & 7) != 0)
		{
			++it;
		}
		else
		{
			it = test.erase(it);
		}
	}

	for (auto _ : state)
	{
		for(std::deque<int>::iterator it = test.begin(); it != test.end(); ++it)
		{
			total += *it;
		}
	}
}



static void benchmark_vector_iterate_and_sum(benchmark::State& state) 
{
	std::vector<int> test;
	
	for (unsigned int counter = 0; counter != 100000; ++counter)
	{
		test.push_back(15);
	}

	for(std::vector<int>::iterator it = test.begin(); it != test.end();)
	{
		if ((xor_rand() & 7) != 0)
		{
			++it;
		}
		else
		{
			it = test.erase(it);
		}
	}

	for (auto _ : state)
	{
		for(std::vector<int>::iterator it = test.begin(); it != test.end(); ++it)
		{
			total += *it;
		}
	}
}


// Register the function as a benchmark
BENCHMARK(benchmark_colony_iterate_and_sum);
BENCHMARK(benchmark_deque_iterate_and_sum);
BENCHMARK(benchmark_vector_iterate_and_sum);


//BENCHMARK_MAIN();

int main (int argc, char ** argv)
{
    benchmark::Initialize (&argc, argv);
    benchmark::RunSpecifiedBenchmarks ();
	printf("Total = %f\n", total);
    return 0;
}
